import logging
from pathlib import Path

from datamaestro.download import Download
from datamaestro.definitions import AbstractDataset

from subprocess import run


class gsync(Download):
    """Google sync call"""

    def __init__(self, varname: str, url: str):
        """Synchronize with Google sync

        Args:
            varname: Variable name
            url: The google sync URL (`gs://...`)
        """
        super().__init__(varname)
        self.url = self.definition["url"]

    def download(self, destination: Path):
        if destination.exists():
            return

        syncpath = destination.parent / ("@%s" % destination.name)
        syncpath.mkdir(exist_ok=True, parents=True)
        logging.info("Synchronizing %s into %s", self.url, syncpath)
        run(["gsutil", "-m", "rsync", self.url, syncpath])



import logging
from typing import Optional

from datamaestro.download import Download


class hf_download(Download):
    """Use Hugging Face to donwload a file"""

    def __init__(
        self,
        varname: str,
        repo_id: str,
        *,
        data_files: Optional[str] = None,
        split: Optional[str] = None
    ):
        """Use

        Args:
            varname: Variable name
            repo_id: The HuggingFace repository ID
        """
        super().__init__(varname)
        self.repo_id = repo_id
        self.data_files = data_files
        self.split = split

    def download(self, force=False):
        try:
            from datasets import load_dataset
        except ModuleNotFoundError:
            logging.error("the datasets library is not installed:")
            logging.error("pip install datasets")
            raise

        self.dataset = load_dataset(self.repo_id, data_files=self.data_files)
        return True

    def prepare(self):
        return {
            "repo_id": self.repo_id,
            "data_files": self.data_files,
            "split": self.split,
        }



# Sphinx extension for datamaestro datasets

from typing import Any, Dict, Tuple
from sphinx.ext.autodoc.mock import mock

from docutils import nodes

from sphinx.application import Sphinx
from sphinx.domains import Domain, ObjType
from sphinx.roles import XRefRole
from sphinx.util.docutils import SphinxDirective
from sphinx.locale import _
from sphinx import addnodes
from sphinx.util.nodes import make_refnode
import datamaestro
from datamaestro.data import AbstractDataset
import logging
from myst_parser.config.main import MdParserConfig
from myst_parser.mdit_to_docutils.base import DocutilsRenderer
from myst_parser.parsers.mdit import create_md_parser


class DatasetNode(nodes.paragraph):
    pass


def to_docutils(source: str):
    parser = create_md_parser(MdParserConfig(), DocutilsRenderer)
    return parser.render(source)


class DatasetsDirective(SphinxDirective):
    def dataset_desc(self, ds: AbstractDataset):
        dm = self.env.get_domain("dm")

        assert isinstance(dm, DatamaestroDomain)
        dm.add_dataset(ds.id)

        # indexnode = addnodes.index(entries=[])
        desc = addnodes.desc()
        desc["domain"] = DatamaestroDomain.name
        desc["objtype"] = desc["desctype"] = "dataset"
        desc["classes"].append(DatamaestroDomain.name)

        signodes = addnodes.desc_signature(ds.id, "", is_multiline=True)
        desc.append(signodes)

        signode = addnodes.desc_signature_line()
        signode += nodes.Text("Dataset ")
        signode += addnodes.desc_name(text=ds.id)
        signode["ids"].append("dataset" + "-" + ds.id)
        signodes.append(signode)

        content = addnodes.desc_content()
        desc.append(content)

        if ds.configtype:
            ctype = ds.configtype
            name = f"{ctype.__module__}.{ctype.__qualname__}"

            te = nodes.paragraph()
            te.append(nodes.Text("Experimaestro type: "))

            p = nodes.paragraph()
            returns = addnodes.desc_returns()
            xref = addnodes.pending_xref(
                "",
                nodes.Text(name),
                refdomain="py",
                reftype="class",
                reftarget=name,
            )
            returns.append(xref)
            p.append(returns)

            content.append(p)

        # node.append(nodes.Text(ds.id))
        if ds.name:
            content.append(
                nodes.paragraph("", "", nodes.strong("", nodes.Text(ds.name)))
            )

        if ds.tags or ds.tasks:
            if ds.tags:
                content.append(
                    nodes.paragraph(
                        "",
                        "",
                        nodes.strong("", nodes.Text("Tags: ")),
                        nodes.Text(", ".join(ds.tags)),
                    )
                )
            if ds.tasks:
                content.append(
                    nodes.paragraph(
                        "",
                        "",
                        nodes.strong("", "Tasks: "),
                        nodes.Text(", ".join(ds.tasks)),
                    )
                )

        if ds.url:
            href = nodes.reference(refuri=ds.url)
            href.append(nodes.Text(ds.url))
            p = nodes.paragraph()
            p.append(nodes.Text("External link: "))
            p.append(href)
            content.append(p)

        if ds.description:
            content.extend(to_docutils(ds.description))

        return desc


class RepositoryDirective(DatasetsDirective):
    """Generates the document for a whole repository"""

    has_content = True
    required_arguments = 1
    optional_arguments = 0

    def run(self):
        (repository_id,) = self.arguments
        with mock(self.config.autodoc_mock_imports):
            repository = datamaestro.Context.instance().repository(
                repository_id
            )  # type: Optional[datamaestro.Repository]
            assert repository is not None

            docnodes = []
            for module in repository.modules():
                section = nodes.section(
                    ids=[f"dm-datasets-{repository_id}-{module.id}"]
                )
                docnodes.append(section)

                section += nodes.title("", nodes.Text(module.title))
                section += nodes.paragraph()
                if module.description:
                    section += to_docutils(module.description).children

                for ds in iter(module):
                    section += self.dataset_desc(ds)

        return docnodes


class DatasetDirective(DatasetsDirective):
    has_content = True
    required_arguments = 1
    optional_arguments = 1

    def run(self):
        # --- Retrieve the datasets
        if len(self.arguments) == 2:
            module_name, repository_name = self.arguments
        else:
            (module_name,) = self.arguments
            repository_name = self.env.config["datamaestro_repository"]

        datasets = None
        with mock(self.config.autodoc_mock_imports):
            for repository in datamaestro.Context.instance().repositories():
                if repository_name is None or repository.id == repository_name:
                    datasets = repository.datasets(module_name)
                    if datasets is not None:
                        break

        assert datasets is not None

        # --- Start documenting

        docnodes = []
        # node.document = self.state.document
        if datasets.description:
            docnodes.extend(to_docutils(datasets.description))

        for ds in datasets:
            docnodes.append(self.dataset_desc(ds))
        return docnodes


class DatamaestroDomain(Domain):
    name = "dm"
    object_types = {
        "dataset": ObjType(_("dataset"), "ds"),
    }
    directives = {
        "repository": RepositoryDirective,
        "datasets": DatasetDirective,
    }
    roles = {"ref": XRefRole()}
    indices = {
        # TODO: Add indices for tags and tasks
    }
    initial_data: Dict[str, Dict[str, Tuple[str, str]]] = {
        "datasets": {},  # fullname -> dataset
        "tags": {},  # tag  -> list of datasets,
        "tasks": {},  # task name -> list of datasets
    }

    def add_dataset(self, dsid):
        self.data["datasets"][dsid] = (self.env.docname, f"dataset-{dsid}")

    def resolve_xref(self, env, fromdocname, builder, typ, target, node, contnode):
        logging.debug("[dm/sphinx] Searching for", target)

        ref = self.data["datasets"].get(target, None)
        if ref:
            docname, targ = ref
            return make_refnode(builder, fromdocname, docname, targ, contnode, targ)
        return None


def setup(app: Sphinx) -> Dict[str, Any]:
    """Setup experimaestro for Sphinx documentation"""

    app.add_domain(DatamaestroDomain)
    app.add_node(DatasetNode)

    app.add_config_value("datamaestro_repository", None, True)

    return {"version": datamaestro.version, "parallel_read_safe": True}



import re
from typing import Iterator, Optional
from functools import cached_property
from attrs import field
import importlib
from omegaconf import OmegaConf
from functools import partial
from attrs import define
from datamaestro import BaseRepository
from datamaestro.definitions import AbstractDataset, DatasetWrapper
from datamaestro.data import Base


re_spec = re.compile(r"""^(\w\.)+:(\w+)""")


@define
class RepositoryDataset:
    ids: list[str]
    """ID(s) of this dataset"""

    entry_point: str = field(validator=re_spec.match)
    """The entry point"""

    title: str
    """The full name of the dataset"""

    description: str
    """Description of the dataset"""

    url: Optional[str]
    """The URL"""

    groups: Optional[list[str]]
    """Groups to which this repository belongs"""


@define
class RepositoryAuthors:
    name: str
    email: str


@define
class RepositoryGroup:
    name: str
    tasks: list[str]
    tags: list[str]


@define
class RepositoryConfiguration:
    namespace: str
    authors: list[RepositoryAuthors]
    description: str
    groups: dict[str, RepositoryGroup]
    datasets: list[RepositoryDataset]


class YAMLDataset(AbstractDataset):
    def __init__(self, repository: "YAMLRepository", information: RepositoryDataset):
        super().__init__(repository)
        self.information = information
        self.id = self.information.ids[0]
        self.aliases = set(self.information.ids)

    @cached_property
    def wrapper(self) -> DatasetWrapper:
        module, func_name = self.information.entry_point.split(":")
        wrapper = getattr(importlib.import_module(module), func_name)
        return wrapper

    def _prepare(self) -> "Base":
        return self.wrapper()

    def download(self, **kwargs):
        return self.wrapper.download(**kwargs)


class YAMLRepository(BaseRepository):
    """YAML-based repository"""

    @property
    def id(self):
        return self.configuration.namespace

    @property
    def name(self):
        return self.configuration.namespace

    @cached_property
    def configuration(self):
        schema = OmegaConf.structured(RepositoryConfiguration)
        with importlib.resources.path(
            self.__class__.__module__, "datamaestro.yaml"
        ) as fp:
            conf = OmegaConf.load(fp)

        conf: RepositoryConfiguration = OmegaConf.merge(schema, conf)
        return conf

    def __iter__(self) -> Iterator["AbstractDataset"]:
        return map(partial(YAMLDataset, self), self.configuration.datasets)



from pathlib import Path
import shutil
from tempfile import NamedTemporaryFile
import yaml


class RegistryEntry:
    def __init__(self, registry, key):
        self.key = key
        self.dicts = []
        _key = ""
        for subkey in self.key.split("."):
            _key = "%s.%s" % (_key, subkey) if _key else subkey
            if _key in registry.content:
                self.dicts.insert(0, registry.content[_key])

    def get(self, key, default):
        for d in self.dicts:
            if key in d:
                return d[key]
        return default

    def __getitem__(self, key):
        for d in self.dicts:
            if key in d:
                return d[key]
        raise KeyError(key)


class Registry:
    def __init__(self, path: Path):
        self.path = path
        self.dirty = False
        self.data = {}
        if path.is_file():
            with path.open("r") as fp:
                for key, value in yaml.load(fp, Loader=yaml.BaseLoader).items():
                    self.data[key] = value

    def __getitem__(self, key):
        return RegistryEntry(self, key)

    def __setitem__(self, key, value):
        self.dirty = True
        self.data[key] = value

    def save(self):
        if self.dirty:
            with NamedTemporaryFile("w", delete=False) as f:
                yaml.dump(self.data, f)
            shutil.move(f.name, self.path)
            self.dirty = False



import logging
from experimaestro import Config
import json
from pathlib import PosixPath, Path
from shutil import rmtree
import shutil
import hashlib
from tqdm import tqdm


class TemporaryDirectory:
    def __init__(self, path: Path):
        self.delete = True
        self.path = path

    def __enter__(self):
        logging.debug("Creating directory %s", self.path)
        self.path.mkdir(parents=True, exist_ok=True)
        return self

    def __exit__(self, type, value, traceback):
        if self.delete:
            rmtree(self.path)


def copyfileobjs(fsrc, fdsts, length=0):
    """copy data from file-like object fsrc to file-like objects fdst

    Based on shutil.copyfileobj
    """
    # Localize variable access to minimize overhead.
    if not length:
        length = shutil.COPY_BUFSIZE
    fsrc_read = fsrc.read
    fdst_writes = [fdst.write for fdst in fdsts]
    while True:
        buf = fsrc_read(length)
        if not buf:
            break
        for fdst_write in fdst_writes:
            fdst_write(buf)


class FileChecker:
    """Checks a file"""

    def check(self, path: Path):
        """Check the given file

        returns true if OK
        """
        with path.open("rb") as fp:
            writer = self.write
            chunk = fp.read(8192)
            while chunk:
                writer(chunk)
                chunk = fp.read(8192)
            self.close()

    @property
    def write(self):
        raise NotImplementedError()

    def close(self):
        raise NotImplementedError()


class HashCheck(FileChecker):
    def __init__(self, hashstr: str, hasherfn=hashlib.md5):
        """Check a file against a hash

        :param hashstr: The HASH value
        :param hasherfn: The hash computer, defaults to hashlib.md5
        """
        self.hashstr = hashstr
        self.hasherfn = hasherfn
        self.hasher = None

    @property
    def write(self):
        """Generates a write-like handle"""
        if self.hasher is None:
            self.hasher = self.hasherfn()

        def write(chunk):
            self.hasher.update(chunk)

        return write

    def close(self):
        s = self.hasher.hexdigest()
        if s != self.hashstr:
            raise IOError(f"Digest do not match (expected {self.hashstr}, got {s})")
        logging.info("Hash check OK [%s]", s)


class CachedFile:
    """Represents a downloaded file that has been cached

    The file is automatically deleted when closed if keep is False
    and used is True
    """

    def __init__(self, path, keep=False, others=[]):
        self.path = path
        self.keep = keep
        self._force_delete = False
        self.others = []

    def __enter__(self):
        return self

    def force_delete(self):
        """Force the file to be deleted (even if an exception was thrown)"""
        self.force_delete = True

    def __exit__(self, exc_type, exc_value, traceback):
        # Avoid removing the file if an exception was thrown
        if not self.force_delete and exc_type is not None:
            logging.info("Keeping cache file %s (exception thrown)", self.path)
            return

        try:
            if not self.keep:
                logging.info("Deleting cache file %s", self.path)
                self.path.unlink()
                for other in self.others:
                    other.unlink()
        except Exception as e:
            logging.warning("Could not delete cached file %s [%s]", self.path, e)


def downloadURL(url: str, path: Path, resume: bool = False, size: int = None):
    import requests

    response = None
    pos = 0
    if path.is_file():
        pos = path.stat().st_size
        if resume and pos > 0:
            logging.warning("Trying to resume download from position %d", pos)
            response = requests.get(
                url, headers={"Range": f"bytes={pos}-"}, stream=True
            )
            if (
                response.status_code != requests.codes.PARTIAL_CONTENT
                or response.headers["Content-Range"] is None
            ):
                # Re-start with new request
                logging.warning("Could not resume download (range request failed)")
                path.unlink()
                response.close()
                response = None
                pos = 0
        else:
            path.unlink()

    if response is None:
        response = requests.get(url, stream=True)

    # Valid response
    assert (
        response.status_code >= 200 and response.status_code < 300
    ), f"Status code is not 2XX ({response.status_code})"

    # Get the total size (or use the provided one)
    total_size = int(response.headers.get("content-length", size or 0))

    # Add the current position to the total size (if resuming)
    if total_size > 0:
        total_size += pos

    CHUNK_SIZE = 1024
    with path.open("ab") as f, tqdm(
        initial=pos, total=total_size, unit_scale=True, unit="B"
    ) as t:
        for data in response.iter_content(chunk_size=CHUNK_SIZE):
            f.write(data)
            t.update(len(data))

        return


def deprecated(message, f):
    from inspect import getframeinfo, stack

    def wrapped(*args, **kwargs):
        caller = getframeinfo(stack()[1][0])
        logging.warning(
            "called at %s:%d - %s" % (caller.filename, caller.lineno, message)
        )
        return f(*args, **kwargs)

    return wrapped


# --- JSON


class JsonContext:
    pass


class BaseJSONEncoder(json.JSONEncoder):
    def __init__(self):
        super().__init__()
        self.context = JsonContext()

    def default(self, o):
        return {
            key: value for key, value in o.__dict__.items() if not key.startswith("__")
        }


class JsonEncoder(BaseJSONEncoder):
    """Default JSON encoder"""

    def default(self, o):
        if isinstance(o, PosixPath):
            return str(o.resolve())

        if isinstance(o, Config):
            return {arg.name: value for arg, value in o.__xpm__.xpmvalues()}

        return super().default(o)


class XPMEncoder(BaseJSONEncoder):
    """Experimaestro encoder"""

    def default(self, o):
        if isinstance(o, PosixPath):
            return {"$type": "path", "$value": str(o.resolve())}

        # Data object
        if isinstance(o, Config):
            m = {arg.name: value for arg, value in o.__xpm__.xpmvalues()}
            # Adds XPM type
            m["$type"] = o.__class__.__getxpmtype__().identifier.name
            return m

        return super().default(o)



# See documentation on https://datamaestro.readthedocs.io

from datamaestro.definitions import datatasks, datatags, dataset
from datamaestro.data import Base


@datatags("tag1", "tag2")
@datatasks("task1", "task2")
@dataset(
    Base,
    url="__URL__",
)
def __IDENTIFIER__():
    """Line description

    Long description
    """
    return {}



import io
import re

from . import Transform


class LineTransformStream(io.RawIOBase):
    """Transform line by line"""

    def __init__(self, fileobj, transform):
        self.current = ""
        self.offset = 1
        self.fileobj = fileobj
        self.transform = transform

    def readable(self):
        return True

    def readnext(self):
        # Read next line and transform
        self.offset = 0
        self.current = None
        while not self.current:
            line = self.fileobj.readline().decode("utf-8")
            if len(line) == 0:
                return None

            self.current = self.transform(line).encode("utf-8")

    def readinto(self, b):
        """Read bytes into a pre-allocated, writable bytes-like object b and
        return the number of bytes read"""
        if self.current is None:
            return 0

        offset = 0
        lb = len(b)
        while lb > 0:
            while self.offset >= len(self.current):
                self.readnext()
                if self.current is None:
                    return offset

            # How many bytes to read from current line
            l = min(lb, len(self.current) - self.offset)

            b[offset : (offset + l)] = self.current[self.offset : (self.offset + l)]
            lb -= l
            offset += l
            self.offset += l

        return offset


class Replace(Transform):
    """Line by line transform"""

    def __init__(self, pattern, replacement):
        self.re = re.compile(pattern)
        self.repl = replacement

    def __call__(self, fileobj):
        return LineTransformStream(fileobj, lambda s: self.re.sub(self.repl, s))


class Filter(Transform):
    """Line by line transform"""

    def __init__(self, pattern):
        self.re = re.compile(pattern)

    def filter(self, line):
        if self.re.search(line):
            return line
        return ""

    def __call__(self, fileobj):
        return LineTransformStream(fileobj, self.filter)



from typing import Optional
from datamaestro.definitions import AbstractDataset, hook


@hook("pre-use")
def useragreement(definition: AbstractDataset, message: str, id: Optional[str] = None):
    """Asks for a user-agreement

    :param definition: The dataset for which the agreement is asked
    :param message: The agreement text
    :param id: The ID of the agreement (default to the dataset ID)
    """
    # Skip agreement when testing
    if definition.context.running_test:
        return

    settings = definition.context.user_settings
    id = id or definition.id
    answer = settings.agreements.get(id, "no")
    if answer == "yes":
        return

    answer = definition.context.ask(
        message + "\nanswer [yes/no] ",
        {"yes": "yes", "y": "yes", "no": "no", "n": "no"},
    )

    if answer == "yes":
        settings.agreements[id] = "yes"
        settings.save()
    else:
        raise ValueError("Agreement not accepted")



from typing import Protocol
from pathlib import Path
from datamaestro import Context
from datamaestro.definitions import DatasetWrapper
from datamaestro.download import Resource


class Downloader(Protocol):
    def __call__(self, context: Context, root: Path, *, force=False):
        pass


class CustomResource(Resource):
    def __init__(self, ds_wrapper: DatasetWrapper, downloader: Downloader):
        self.ds_wrapper = ds_wrapper
        self.downloader = downloader

    def prepare(self):
        pass

    def download(self, force=False):
        self.downloader(self.context, self.ds_wrapper.datapath, force=force)


def custom_download(downloader: Downloader) -> Path:
    ds_wrapper = DatasetWrapper.BUILDING[-1]
    ds_wrapper.ordered_resources.append(CustomResource(ds_wrapper, downloader))

    return ds_wrapper.datapath



import logging
import os
from datamaestro.utils import deprecated
from datamaestro.definitions import AbstractDataset
from typing import List
from datamaestro.download import Download
from datamaestro.context import ResolvablePath
from pathlib import Path
import os
import logging


class links(Download):
    def __init__(self, varname: str, **links: List[AbstractDataset]):
        """Link with another dataset path

        Args:
            varname: The name of the variable when defining the dataset
            links: A list of
        """
        super().__init__(varname)
        self.links = links

    @property
    def path(self):
        return self.definition.datapath

    def prepare(self):
        return self.path

    def download(self, force=False):
        self.path.mkdir(exist_ok=True, parents=True)
        for key, value in self.links.items():
            value.download(force)

            path = value()
            dest = self.path / key

            if not dest.exists():
                if dest.is_symlink():
                    logging.info("Removing dandling symlink %s", dest)
                    dest.unlink()
                os.symlink(path, dest)


# Deprecated
Links = deprecated("Use @links instead of @Links", links)


class linkpath(Download):
    def __init__(self, varname: str, proposals):
        """Link to a folder

        Args:
            varname: Name of the variable
            proposals: List of potential paths
        """
        super().__init__(varname)
        self.proposals = proposals

    def prepare(self):
        return self.path

    @property
    def path(self):
        return self.definition.datapath / self.varname

    def download(self, destination):
        if self.check(self.path):
            return

        if self.path.is_symlink():
            logging.warning("Removing dandling symlink %s", self.path)
            self.path.unlink()

        path = None

        for searchpath in self.proposals:
            logging.info("Trying path %s", searchpath)
            try:
                path = ResolvablePath.resolve(self.context, searchpath)
                if self.check(path):
                    break
                logging.info("Path %s not found", path)
            except KeyError:
                logging.info("Could not expand path %s", searchpath)

        # Ask the user
        while path is None or not self.check(path):
            path = Path(input("Path to %s: " % self.varname))
        assert path.name

        logging.debug("Linking %s to %s", path, self.path)
        self.path.parent.mkdir(exist_ok=True, parents=True)
        os.symlink(path, self.path)


class linkfolder(linkpath):
    def check(self, path):
        return path.is_dir()

    def __init__(self, varname: str, proposals):
        """Link to a folder

        Args:
            varname: Name of the variable
            proposals: List of potential paths
        """
        super().__init__(varname, proposals)


class linkfile(linkpath):
    def __init__(self, varname: str, proposals):
        """Link to a file

        Args:
            varname: Name of the variable
            proposals: List of potential paths
        """
        super().__init__(varname, proposals)

    def check(self, path):
        print("Checking", path, path.is_file())
        return path.is_file()



#
# Main datamaestro functions and data models
#

import logging
import inspect
from pathlib import Path
from itertools import chain
from abc import ABC, abstractmethod
from contextlib import contextmanager
import traceback
from typing import (
    Dict,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Callable,
    TYPE_CHECKING,
    Union,
    ClassVar,
    _GenericAlias,
)
from experimaestro import (  # noqa: F401 (re-exports)
    argument,
    constant,
    Param,
    Option,
    Config,
    Meta,
)
from typing import Type as TypingType  # noqa: F401 (re-exports)
from experimaestro.core.types import Type  # noqa: F401 (re-exports)

if TYPE_CHECKING:
    from .data import Base, Dataset
    from .context import Repository, Context, DatafolderPath  # noqa: F401 (re-exports)
    from datamaestro.download import Download

# --- Objects holding information into classes/function


class AbstractData(ABC):
    """Data definition groups common fields between a dataset and a data piece,
    such as tags and tasks"""

    def __init__(self):
        self.tags = set()
        self.tasks = set()

    def ancestors(self):
        """Returns all configuration ancestors"""
        ancestors = []
        ancestors.extend(c for c in self.base.__mro__ if issubclass(c, Config))
        return ancestors


class DataDefinition(AbstractData):
    """Object that stores the declarative part of a data(set) description"""

    def __init__(self, t, base=None):
        assert base is None or not inspect.isclass(t)

        super().__init__()

        # Copy base type and find matching repository
        self.t = t

        # Dataset id (and all aliases)
        self.id = None
        self.base = base

        self.aliases: Set[str] = set()

        self.tags = set(chain(*[c.__datamaestro__.tags for c in self.ancestors()]))
        self.tasks = set(chain(*[c.__datamaestro__.tasks for c in self.ancestors()]))
        self._description: Optional[str] = None

    @property
    def description(self):
        return self._description

    @staticmethod
    def repository_relpath(t: type) -> Tuple["Repository", List[str]]:
        """Find the repository of the current data or dataset definition"""
        from .context import Context  # noqa: F811

        repositorymap = Context.instance().repositorymap

        fullname = f"{t.__module__}.{t.__name__}"
        components = fullname.split(".")

        current: str = ""
        longest_ix = -1
        repository = None
        for ix, c in enumerate(components):
            current = f"{current}.{c}" if current else c
            if (current is not None) and (current in repositorymap):
                longest_ix = ix
                repository = repositorymap[current]

        if repository is None:
            if components[0] == "datamaestro":
                longest_ix = 0

        return repository, components[(longest_ix + 1) :]

    def ancestors(self):
        ancestors = []
        if self.base:
            baseclass = self.base
        else:
            baseclass = self.t

        ancestors.extend(c for c in baseclass.__mro__ if hasattr(c, "__datamaestro__"))

        return ancestors


class AbstractDataset(AbstractData):
    """Specialization of AbstractData for datasets

    A dataset:

    - has a unique ID (and aliases)
    - can be searched for
    - has a data storage space
    - has specific attributes:
        - timestamp: whether the dataset version depends on the time of the download
    """

    name: Optional[str] = None
    """The name of the dataset"""

    url: Optional[str] = None
    """The URL of the dataset"""

    doi: Optional[str] = None
    """The DOI of this dataset"""

    def __init__(self, repository: Optional["Repository"]):
        super().__init__()
        self.repository = repository
        self.timestamp = False
        self.aliases = set()

        # Associated resources
        self.resources: Dict[str, "Download"] = {}
        self.ordered_resources = []

        # Hooks
        # pre-use: before returning the dataset object
        # pre-download: before downloading the dataset
        self.hooks = {"pre-use": [], "pre-download": []}

        self.url = None
        self.version = None

    @property
    def description(self):
        raise NotImplementedError(f"For class {self.__class__}")

    @property
    def configtype(self):
        raise NotImplementedError()

    @property
    def context(self):
        if self.repository is None:
            from datamaestro.context import Context  # noqa: F811

            return Context.instance()
        return self.repository.context

    def prepare(self, download=False) -> "Base":
        ds = self._prepare()
        ds.__datamaestro_dataset__ = self

        if download:
            ds.download()
        return ds

    def register_hook(self, hookname: str, hook: Callable):
        self.hooks[hookname].append(hook)

    @abstractmethod
    def _prepare(self) -> "Base":
        ...

    def format(self, encoder: str) -> str:
        s = self.prepare()
        if encoder == "normal":
            from .utils import JsonEncoder

            return JsonEncoder().encode(s)
        elif encoder == "xpm":
            from .utils import XPMEncoder

            return XPMEncoder().encode(s)
        else:
            raise Exception("Unhandled encoder: {encoder}")

    def setDataIDs(self, data: Config, id: str):
        """Set nested IDs automatically"""
        from datamaestro.data import Base

        if isinstance(data, Base):
            if self.repository is None:
                data.id = id
            else:
                data.id = f"{id}@{self.repository.name}"
        for key, value in data.__xpm__.values.items():
            if isinstance(value, Config):
                self.setDataIDs(value, f"{id}.{key}")

    def download(self, force=False):
        """Download all the necessary resources"""
        success = True
        logging.info("Materializing %d resources", len(self.ordered_resources))
        self.prepare()
        for resource in self.ordered_resources:
            try:
                resource.download(force)
            except Exception:
                logging.error("Could not download resource %s", resource)
                traceback.print_exc()
                success = False
                break
        return success

    @staticmethod
    def find(name: str) -> "DataDefinition":
        """Find a dataset given its name"""
        from datamaestro.context import Context  # noqa: F811

        logging.debug("Searching dataset %s", name)
        for repository in Context.instance().repositories():
            logging.debug("Searching dataset %s in %s", name, repository)
            dataset = repository.search(name)
            if dataset is not None:
                return dataset
        raise Exception("Could not find the dataset %s" % (name))


class FutureAttr:
    """Allows to access a dataset sub-property"""

    def __init__(self, dataset, keys):
        self.dataset = dataset
        self.keys = keys

    def __repr__(self):
        return "[%s].%s" % (self.dataset.id, ".".join(self.keys))

    def __call__(self):
        """Returns the value"""
        value = self.dataset.prepare()
        for key in self.keys:
            value = getattr(value, key)
        return value

    def __getattr__(self, key):
        return FutureAttr(self.dataset, self.keys + [key])

    def download(self, force=False):
        self.dataset.download(force)


class DatasetWrapper(AbstractDataset):
    """Wraps an annotated method into a dataset

    This is the standard way to define a dataset in datamaestro through
    annotations (otherwise, derive from `AbstractDataset`).
    """

    BUILDING: ClassVar[list["DatasetWrapper"]] = []
    """Currently built dataset"""

    def __init__(self, annotation, t: type):
        self.config = None
        self.repository: Optional[Repository] = None
        self.t = t
        self.base = annotation.base
        assert self.base is not None, f"Could not set the Config type for {t}"

        repository, components = DataDefinition.repository_relpath(t)
        super().__init__(repository)

        # Set some variables
        self.url = annotation.url
        self.doi = annotation.doi

        # Builds the ID:
        # Removes module_name.config prefix
        if annotation.id is None or annotation.id == "":
            # Computes an ID
            assert (
                # id is empty string = use the module id
                components[0]
                == "config"
            ), (
                "A @dataset without `id` should be in the "
                f".config module (not {t.__module__})"
            )
            path = ".".join(components[1:-1])

            self.id = path
        else:
            # Use the provided ID
            self.id = annotation.id

        self.aliases.add(self.id)

        # Get the documentation
        self._name = None
        self._description = None

    @property
    def name(self):
        self._process_doc()
        return self._name

    @property
    def description(self):
        self._process_doc()
        return self._description

    def _process_doc(self):
        if self._description is None:
            if self.t.__doc__:
                lines = self.t.__doc__.split("\n")
                self._name = lines[0]
                if len(lines) > 1:
                    assert lines[1].strip() == "", "Second line should be blank"
                if len(lines) > 2:
                    # Remove the common indent
                    lines = [line.rstrip() for line in lines[2:]]
                    minindent = max(
                        next(idx for idx, chr in enumerate(s) if not chr.isspace())
                        for s in lines
                        if len(s) > 0
                    )
                    self._description = "\n".join(
                        s[minindent:] if len(s) > 0 else "" for s in lines
                    )
            else:
                self._name = ""
                self._description = ""

    @property
    def configtype(self):
        return self.base

    def __getattr__(self, key):
        """Returns a pointer to a potential attribute"""
        return FutureAttr(self, [key])

    def download(self, force=False):
        if self.base is self.t:
            self._prepare()
        return super().download(force=force)

    @contextmanager
    def building(self):
        DatasetWrapper.BUILDING.append(self)
        yield self
        DatasetWrapper.BUILDING.pop()

    def _prepare(self) -> "Base":
        if self.config is not None:
            return self.config

        # Direct creation of the dataset
        if self.base is self.t:
            self.config = self.base.__create_dataset__(self)

        # Construct the object
        resources = {key: value.prepare() for key, value in self.resources.items()}

        with self.building():
            result = self.t(**resources)

        # Download resources
        logging.debug("Building with data type %s and dataset %s", self.base, self.t)
        for hook in self.hooks["pre-use"]:
            hook(self)

        if result is None:
            name = self.t.__name__
            filename = inspect.getfile(self.t)
            raise Exception(
                f"The dataset method {name} defined in "
                f"{filename} returned a null object"
            )

        if isinstance(result, dict):
            self.config = self.base(**result)
        elif isinstance(result, self.base):
            self.config = result
        else:
            raise RuntimeError(
                f"The dataset method {name} defined in "
                f"{filename} returned an object of type {type(dict)}"
            )

        # Setup ourself
        self.config.__datamaestro_dataset__ = self

        # Set the ids
        self.setDataIDs(self.config, self.id)

        return self.config

    __call__ = _prepare

    @property
    def _path(self) -> Path:
        """Returns a unique relative path for this dataset"""
        path = Path(*self.id.split("."))
        if self.version:
            path = path.with_suffix(".v%s" % self.version)
        return path

    @property
    def datapath(self):
        """Returns the destination path for downloads"""
        from datamaestro import Context  # noqa: F811

        path = Context.instance().storepath / self._path

        if (self.repository is not None) and (not path.exists()):
            old_path: Path = self.repository.datapath / self._path
            if old_path.exists():
                logging.info(
                    "Moving from old path [%s] to new path [%s]", old_path, path
                )
                path.parent.mkdir(exist_ok=True, parents=True)
                old_path.rename(path)

        return path

    def hasfiles(self) -> bool:
        """Returns whether this dataset has files or only includes references"""
        for resource in self.resources.values():
            if resource.hasfiles():
                return True

        return False


# --- Annotations

T = TypeVar("T")


class DataAnnotation:
    def __call__(self, object: T) -> T:
        if isinstance(object, AbstractDataset):
            self.annotate(object)
        else:
            if "__datamaestro__" in object.__dict__:
                self.annotate(object.__datamaestro__)
            else:
                # With configuration objects, add a __datamaestro__ member to the class
                assert issubclass(
                    object, Config
                ), f"{object} cannot be annotated (only dataset or data definitions)"
                if "__datamaestro__" not in object.__dict__:
                    object.__datamaestro__ = AbstractData()
                self.annotate(object.__datamaestro__)

        return object

    def annotate(self, data: AbstractData):
        raise NotImplementedError("Method annotate for class %s" % self.__class__)


class DatasetAnnotation:
    """Base class for all annotations"""

    def __call__(self, dataset: AbstractDataset):
        if isinstance(dataset, AbstractDataset):
            self.annotate(dataset)
        elif issubclass(dataset, Dataset):
            self.annotate(dataset.__datamaestro__)
        else:
            raise RuntimeError(
                f"Only datasets can be annotated with {self}, "
                f"but {dataset} is not a dataset"
            )

        return dataset

    def annotate(self, dataset: AbstractDataset):
        raise NotImplementedError("Method annotate for class %s" % self.__class__)


def hook(name: str):
    """Annotate a method of a DatasetAnnotation class to be a hook"""

    class HookAnnotation(DatasetAnnotation):
        def __init__(self, callable: Callable, args, kwargs):
            self.callable = callable
            self.args = args
            self.kwargs = kwargs

        def annotate(self, dataset):
            dataset.register_hook(name, self._hook)

        def _hook(self, definition: DataDefinition):
            self.callable(definition, *self.args, **self.kwargs)

    def annotate(callable: Callable):
        def collect(*args, **kwargs):
            return HookAnnotation(callable, args, kwargs)

        return collect

    return annotate


def DataTagging(f):
    class Annotation(DataAnnotation):
        """Define tags in a data definition"""

        def __init__(self, *tags):
            self.tags = tags

        def annotate(self, metadata):
            f(metadata).update(self.tags)

    return Annotation


datatags = DataTagging(lambda d: d.tags)
datatasks = DataTagging(lambda d: d.tasks)


class metadata:
    def __init__(
        self, tags: Union[str, List[str]] = None, tasks: Union[str, List[str]] = None
    ):
        pass

    def __call__(self, object: type):
        # FIXME: todo
        return object


class dataset:
    def __init__(
        self,
        base=None,
        *,
        timestamp=None,
        id=None,
        url=None,
        size=None,
        doi=None,
    ):
        """Creates a new (meta)dataset

        Meta-datasets are not associated with any base type

        Arguments:
            base {[type]} -- The base type (or None if infered from type
            annotation)

        Keyword Arguments:
            timestamp {bool} -- If the dataset evolves, specify its timestamp
            (default: None)

            id {[type]} -- [description] (default: {None})

            url {[type]} -- [description] (default: {None})

            size {str} -- The size (should be a parsable format)

            doi {str} -- The DOI of the corresponding paper
        """
        if hasattr(base, "__datamaestro__") and isinstance(
            base.__datamaestro__, metadataset
        ):
            self.base = base.__datamaestro__.base
        else:
            self.base = base

        self.id = id
        self.url = url
        self.meta = False
        self.timestamp = timestamp
        self.size = size
        self.doi = doi

    def __call__(self, t):
        try:
            if self.base is None:
                from datamaestro.data import Base

                if inspect.isclass(t) and issubclass(t, Base):
                    self.base = t
                else:
                    try:
                        # Get type from return annotation
                        return_type = t.__annotations__["return"]
                        if isinstance(return_type, _GenericAlias):
                            return_type = return_type.__origin__
                        self.base = return_type
                    except KeyError:
                        logging.warning("No return annotation in %s", t)
                        raise
            object.__getattribute__(t, "__datamaestro__")
            raise AssertionError("@data should only be called once")
        except AttributeError:
            pass
        dw = DatasetWrapper(self, t)
        t.__dataset__ = dw
        if inspect.isclass(t) and issubclass(t, Base):
            return t
        return dw


class metadataset(AbstractDataset):
    """Annotation for object/functions which are abstract dataset definitions

    i.e. shared by more than one real dataset. This is useful to share tags,
    urls, etc.
    """

    def __init__(self, base):
        super().__init__(None)
        self.base = base

    def __call__(self, t):
        try:
            object.__getattribute__(t, "__datamaestro__")
            raise AssertionError("@data should only be called once")
        except AttributeError:
            pass
        t.__datamaestro__ = self
        return t

    _prepare = None



from typing import Optional
import logging
import shutil
import tarfile
import io
import gzip
import os.path as op
import os
import urllib3
from pathlib import Path
import re
from datamaestro.utils import copyfileobjs, FileChecker
from datamaestro.stream import Transform
from datamaestro.download import Download


def open_ext(*args, **kwargs):
    """Opens a file according to its extension"""
    name = args[0]
    if name.endswith(".gz"):
        return gzip.open(*args, *kwargs)
    return io.open(*args, **kwargs)


class SingleDownload(Download):
    def __init__(self, filename: str):
        super().__init__(re.sub(r"\..*$", "", filename))
        self.name = filename

    @property
    def path(self):
        return self.definition.datapath / self.name

    def prepare(self):
        return self.path

    def download(self, force=False):
        if not self.path.is_file():
            self._download(self.path)


class filedownloader(SingleDownload):
    def __init__(
        self,
        filename: str,
        url: str,
        size: int = None,
        transforms: Optional[Transform] = None,
        checker=None,
    ):
        """Downloads a file given by a URL

        :param filename: The filename within the data folder; the variable name
            corresponds to the filename without the extension.

        :param url: The URL to download.

        :param transforms: Transform the file before storing it size: size in
            bytes (or None)
        """
        super().__init__(filename)
        self.url = url
        self.checker = checker
        self.size = size

        p = urllib3.util.parse_url(self.url)
        path = Path(Path(p.path).name)
        self.transforms = transforms if transforms else Transform.createFromPath(path)

    def _download(self, destination):
        logging.info("Downloading %s into %s", self.url, destination)

        # Creates directory if needed
        dir = op.dirname(destination)
        os.makedirs(dir, exist_ok=True)

        # Download (cache)
        with self.context.downloadURL(self.url, size=self.size) as file:
            # Transform if need be
            if self.transforms:
                logging.info("Transforming file")
                with self.transforms(file.path.open("rb")) as stream, destination.open(
                    "wb"
                ) as out:
                    if self.checker:
                        copyfileobjs(stream, [out, self.checker])
                        self.checker.close()
                    else:
                        shutil.copyfileobj(stream, out)
            else:
                logging.info("Keeping original downloaded file %s", file.path)
                if self.checker:
                    self.checker.check(file.path)
                (shutil.copy if file.keep else shutil.move)(file.path, destination)

        logging.info("Created file %s" % destination)


def file_from_url(
    filename: str,
    url: str,
    *,
    size: Optional[int] = None,
    transforms: Optional[Transform] = None,
    checker: Optional[FileChecker] = None,
) -> Path:
    """Defines a file that should be downloaded from"""
    downloader = filedownloader(filename, url, size, transforms, checker)
    downloader.contextualize()
    return downloader.path


class concatdownload(SingleDownload):
    """Concatenate all files in an archive"""

    def __init__(self, filename: str, url: str, transforms=None):
        """Concat the files in an archive

        Args:
            filename: The filename within the data folder; the variable name
            corresponds to the filename without the extension url: The URL to
            download transforms: Transform the file before storing it
        """
        super().__init__(filename)
        self.url = url
        self.transforms = transforms

    def _download(self, destination):
        with self.context.downloadURL(self.url) as dl, tarfile.open(dl.path) as archive:
            destination.parent.mkdir(parents=True, exist_ok=True)

            with open(destination, "wb") as out:
                for tarinfo in archive:
                    if tarinfo.isreg():
                        transforms = self.transforms or Transform.createFromPath(
                            Path(tarinfo.name)
                        )
                        logging.debug("Processing file %s", tarinfo.name)
                        with transforms(archive.fileobject(archive, tarinfo)) as fp:
                            shutil.copyfileobj(fp, out)



from pathlib import Path

from datamaestro.download import Download


class Todo(Download):
    def download(self, destination: Path):
        raise NotImplementedError(
            "Download method not defined - please edit the definition file"
        )



"""Global and user settings utility classes"""
import marshmallow as mm
from typing import Dict, Any
from experimaestro.utils.settings import JsonSettings, PathField
from pathlib import Path

# --- Global settings


class SettingsSchema(mm.Schema):
    keys = mm.fields.Dict(keys=mm.fields.Str(), values=mm.fields.Str())
    datafolders = mm.fields.Dict(keys=mm.fields.Str(), values=mm.fields.Str())

    @mm.post_load
    def make_settings(self, data, **kwargs):
        settings = Settings()
        for key, value in data.items():
            setattr(settings, key, value)
        return settings


class Settings(JsonSettings):
    SCHEMA = SettingsSchema

    """Global settings"""

    def __init__(self):
        self.keys: Dict[str, Any] = {}
        self.datafolders: Dict[str, Path] = {}


# --- User settings


class UserSettingsSchema(mm.Schema):
    agreements = mm.fields.Dict(keys=mm.fields.Str(), values=mm.fields.Str())

    @mm.post_load
    def make_settings(self, data, **kwargs):
        settings = UserSettings()
        for key, value in data.items():
            setattr(settings, key, value)
        return settings


class UserSettings(JsonSettings):
    """User settings"""

    SCHEMA = UserSettingsSchema

    def __init__(self):
        self.agreements: Dict[str, str] = {}



from typing import Optional, Tuple, List, Any
from csv import reader as csv_reader
from experimaestro import Param, Meta
from experimaestro import documentation
from . import File


class Generic(File):
    """A generic CSV file"""

    delimiter: Meta[str] = ","
    ignore: Meta[int] = 0
    names_row: Meta[int] = -1

    @documentation
    def columns(self):
        """Returns the list of field names (if any) or None"""
        if self.names_row < 0:
            return None

        with self.path.open("r") as fp:
            for i in range(self.ignore):
                fp.readline()

            for ix, row in enumerate(csv_reader(fp, delimiter=self.delimiter)):
                if ix == self.names_row:
                    return row


class Matrix(Generic):
    """A numerical dataset"""

    names_row: Param[int] = -1
    size_row: Param[int] = -1
    target: Param[Optional[str]] = None

    @documentation
    def data(self) -> Tuple[List[str], Any]:
        """Returns the list of fields and the numeric data


        Returns: List of fields
        """
        import numpy as np

        fields = []
        targets = None if self.size_row >= 0 or not self.target else []
        data = None if self.size_row >= 0 else []
        i = 0
        skipix = -1

        with self.path.open("r") as fp:
            for i in range(self.ignore):
                fp.readline()

            for ix, row in enumerate(csv_reader(fp)):
                if ix == self.size_row:
                    data = np.empty((int(row[0]), int(row[1])))
                elif ix == self.names_row:
                    fields = row
                    if self.target:
                        skipix = fields.index(self.target)
                else:
                    if self.size_row < 0:
                        data.append(
                            [float(x) for ix, x in enumerate(row) if ix != skipix]
                        )
                        if skipix >= 0:
                            targets.append(float(row[skipix]))
                    else:
                        data[i] = [float(x) for ix, x in enumerate(row) if ix != skipix]
                        if skipix >= 0:
                            targets[i] = float(row[skipix])
                        i += 1

        if self.size_row < 0:
            data = np.array(data)
            if targets:
                targets = np.array(targets)

        if self.target:
            return fields, data, targets

        return fields, data



from typing import Union
from abc import ABC, abstractmethod
from datamaestro.definitions import AbstractDataset, DatasetAnnotation
from datamaestro.utils import deprecated
from attrs import define


def initialized(method):
    """Ensure the object is initialized"""

    def wrapper(self, *args, **kwargs):
        if not self._post:
            self._post = True
            self.postinit()
        return method(self, *args, **kwargs)

    return wrapper


@define(kw_only=True)
class SetupOptions:
    pass


class Resource(DatasetAnnotation, ABC):
    """
    Base class for all download handlers
    """

    def __init__(self, varname: str):
        self.varname = varname
        # Ensures that the object is initialized
        self._post = False
        self.definition = None

    def annotate(self, dataset: AbstractDataset):
        assert self.definition is None
        # Register has a resource download
        if self.varname in dataset.resources:
            raise AssertionError("Name %s already declared as a resource", self.varname)

        dataset.resources[self.varname] = self
        dataset.ordered_resources.append(self)
        self.definition = dataset

    def contextualize(self):
        """When using an annotation inline, uses the current dataset wrapper object"""
        from datamaestro.definitions import DatasetWrapper

        wrapper = DatasetWrapper.BUILDING[-1]
        self.annotate(wrapper)

    @property
    def context(self):
        return self.definition.context

    def postinit(self):
        pass

    def hasfiles(self):
        return True

    @abstractmethod
    def download(self, force=False):
        """Downloads the content"""
        ...

    @abstractmethod
    def prepare(self):
        """Prepares the dataset"""
        ...

    def setup(
        self,
        dataset: Union[AbstractDataset],
        options: SetupOptions = None,
    ):
        """Direct way to setup the resource (no annotation)"""
        self(dataset)
        return self.prepare()


# Keeps downwards compatibility
Download = Resource


class reference(Resource):
    def __init__(self, varname=None, reference=None):
        """References another dataset

        :param varname: The name of the variable
        :param reference: Another dataset
        """
        super().__init__(varname)
        assert reference is not None, "Reference cannot be null"
        self.reference = reference

    def prepare(self):
        v = self.reference.prepare()
        if isinstance(v, AbstractDataset):
            return v().prepare()
        return v

    def download(self, force=False):
        self.reference.__datamaestro__.download(force)

    def hasfiles(self):
        # We don't really have files
        return False


Reference = deprecated("Use @reference instead of @Reference", reference)



import pickle
from datamaestro.record import Item, record_type
from attrs import define
import pytest


@define
class AItem(Item):
    a: int


@define
class A1Item(AItem):
    a1: int


@define
class BItem(Item):
    b: int


@define
class B1Item(BItem):
    b1: int


@define
class CItem(Item):
    c: int


ARecord = record_type(AItem)
BaseRecord = ARecord.sub(A1Item)
MyRecord = BaseRecord.sub(BItem)


def test_record_simple():
    a = A1Item(1, 2)
    b = BItem(4)
    r = MyRecord(a, b)
    assert r[AItem] is a
    assert r[A1Item] is a
    assert r[BItem] is b


def test_record_missing_init():
    with pytest.raises(KeyError):
        # A1Item is missing
        MyRecord(AItem(1), BItem(2))

    with pytest.raises(KeyError):
        MyRecord(A1Item(1, 2))


def test_record_update():
    a = A1Item(1, 2)
    b = BItem(4)
    r = MyRecord(a, b)

    r2 = r.update(BItem(3))
    assert r is not r2
    assert r2[BItem] is not b


def test_record_pickled():
    # First,
    MyRecord2 = BaseRecord.sub(BItem)
    r = MyRecord2(A1Item(1, 2), BItem(2))
    r = pickle.loads(pickle.dumps(r))

    assert r[A1Item].a == 1
    assert r[BItem].b == 2



import logging
import importlib
import inspect
from datamaestro.context import Context


class DatasetTests:
    @classmethod
    def setUpClass(cls):
        context = Context.instance()
        module = importlib.import_module(cls.__module__.split(".")[0])
        logging.info("Setting up %s", module.Repository(context))
        cls.__DATAMAESTRO_REPOSITORY__ = module.Repository(context)

    @property
    def repository(self):
        return self.__class__.__DATAMAESTRO_REPOSITORY__

    def test_datafiles(self):
        for context, file_id, package in self.repository._modules():
            with self.subTest(package=package):
                importlib.import_module(package)

    def test_unique_id(self):
        """Test if IDs are unique within the module"""
        mapping = {}
        for dataset in self.repository:
            for id in dataset.aliases:
                mapping.setdefault(id, []).append(dataset.t)

        flag = True
        for key, values in mapping.items():
            if len(values) > 1:
                flag = False
                logging.error("Id %s has several mappings", key)
                for value in values:
                    filename = inspect.getfile(value)
                    lineno = inspect.getsourcelines(value)[1]
                    logging.error("%s:%d", filename, lineno)

        assert flag

    def test_datasets(self):
        """Check datasets integrity by preparing them (without downloading)

        Arguments:
            repository {Repository} -- The repository to check
        """
        for dataset in self.repository:
            with self.subTest(dataset_id=dataset.id):
                dataset.prepare(download=False)



import logging
from pathlib import Path
import zipfile
import shutil
import urllib3
import tarfile
import re
from typing import Set
from datamaestro.download import Download, initialized
from datamaestro.utils import CachedFile, FileChecker


class ArchiveDownloader(Download):
    """Abstract class for all archive related extractors"""

    def __init__(
        self,
        varname,
        url: str,
        subpath: str = None,
        checker: FileChecker = None,
        files: Set[str] = None,
    ):
        """Downloads and extract the content of the archive

        Args:
            varname: The name of the variable when defining the dataset

            url: The archive URL

            checker: the hash check for the downloaded file, composed of two

            subpath: A subpath in the archive; only files from this subpath will
            be extracted

            files: A set of files; if present, only extract those
        """
        super().__init__(varname)
        self.url = url
        self.subpath = subpath
        self.checker = checker
        self._files = set(files) if files else None
        if self.subpath and not self.subpath.endswith("/"):
            self.subpath = self.subpath + "/"

    def postinit(self):
        # Define the path
        p = urllib3.util.parse_url(self.url)
        name = self._name(Path(p.path).name)

        if len(self.definition.resources) > 1:
            self.path = self.definition.datapath / name
        else:
            self.path = self.definition.datapath

    @initialized
    def prepare(self):
        return self.path

    @property
    def extractall(self):
        """Returns whether everything can be extracted"""
        return self._files is None and self.subpath is None

    def filter(self, iterable, getname):
        L = len(self.subpath) if self.subpath else 0

        for info in iterable:
            name = getname(info)
            logging.debug("Looking at %s", name)
            if self._files and not (name in self._files):
                continue

            if self.subpath and name.startswith(self.subpath):
                yield info, name[L:]

            if not self.subpath:
                yield info, name

    @initialized
    def download(self, force=False):
        # Already downloaded
        destination = self.definition.datapath
        if destination.is_dir():
            return

        logging.info("Downloading %s into %s", self.url, destination)

        destination.parent.mkdir(parents=True, exist_ok=True)
        tmpdestination = destination.with_suffix(".tmp")
        if tmpdestination.exists():
            logging.warn("Removing temporary directory %s", tmpdestination)
            shutil.rmtree(tmpdestination)

        with self.context.downloadURL(self.url) as file:
            if self.checker:
                self.checker.check(file.path)
            self.unarchive(file, tmpdestination)

        # Look at the content
        for ix, path in enumerate(tmpdestination.iterdir()):
            if ix > 1:
                break

        # Just one folder: move
        if ix == 0 and path.is_dir():
            logging.info(
                "Moving single file/directory {} into destination {}".format(
                    path, destination
                )
            )
            shutil.move(str(path), str(destination))
            shutil.rmtree(tmpdestination)
        else:
            shutil.move(tmpdestination, destination)


class zipdownloader(ArchiveDownloader):
    """ZIP Archive handler"""

    def _name(self, name):
        return re.sub(r"\.zip$", "", name)

    def unarchive(self, file, destination: Path):
        logging.info("Unzipping file")
        with zipfile.ZipFile(file.path) as zip:
            if self.extractall:
                zip.extractall(destination)
            else:
                for zip_info, name in self.filter(
                    zip.infolist(), lambda zip_info: zip_info.filename
                ):
                    if zip_info.is_dir():
                        (destination / name).mkdir()
                    else:
                        logging.info(
                            "File %s (%s) to %s",
                            zip_info.filename,
                            name,
                            destination / name,
                        )
                        with zip.open(zip_info) as fp, (destination / name).open(
                            "wb"
                        ) as out:
                            shutil.copyfileobj(fp, out)


class tardownloader(ArchiveDownloader):
    """TAR archive handler"""

    def _name(self, name):
        return re.sub(r"\.tar(\.gz|\.bz\|xz)?$", "", name)

    def unarchive(self, file: CachedFile, destination: Path):
        logging.info("Unarchiving file")
        if self.subpath:
            raise NotImplementedError()

        with tarfile.TarFile.open(file.path) as tar:
            if self.extractall:
                tar.extractall(destination)
            else:
                for info, name in self.filter(tar, lambda info: info.name):
                    if info.isdir():
                        (destination / name).mkdir()
                    else:
                        logging.info(
                            "File %s (%s) to %s",
                            info.name,
                            name,
                            destination / name,
                        )
                        logging.info("Extracting into %s", destination / name)
                        tar.extract(info, destination / name)



from .definitions import AbstractDataset
import re


class Condition:
    PATTERNS = [
        (re.compile(r"^tag:(.*)"), lambda m: TagCondition(m.group(1))),
        (re.compile(r"^task:(.*)"), lambda m: TaskCondition(m.group(1))),
        (re.compile(r"^type:(.*)"), lambda m: TypeCondition(m.group(1))),
        (
            re.compile(r"^repo(?:sitory)?:(.*)"),
            lambda m: RepositoryCondition(m.group(1)),
        ),
    ]

    def match(self, dataset: AbstractDataset):
        raise Exception("Match not implemented in %s" % type(self))

    @staticmethod
    def parse(searchterm):
        for regex, callback in Condition.PATTERNS:
            m = regex.match(searchterm)
            if m:
                return callback(m)

        return IDCondition(searchterm)


class AndCondition(Condition):
    def __init__(self):
        self.conditions = []

    def append(self, condition: Condition):
        self.conditions.append(condition)

    def match(self, dataset: AbstractDataset):
        for condition in self.conditions:
            if not condition.match(dataset):
                return False
        return True

    def __repr__(self):
        return " AND ".join([repr(x) for x in self.conditions])


class OrCondition(Condition):
    def __init__(self):
        self.conditions = []

    def append(self, condition: Condition):
        self.conditions.append(condition)

    def match(self, dataset: AbstractDataset):
        for condition in self.conditions:
            if condition.match(dataset):
                return True
        return False

    def __repr__(self):
        return " OR ".join(repr(s) for s in self.conditions)


class ReCondition(Condition):
    def __init__(self, regex):
        self.regex = re.compile(regex)


class TagCondition(ReCondition):
    def match(self, dataset: AbstractDataset):
        for tag in dataset.tags:
            if self.regex.search(tag):
                return True
        return False


class TaskCondition(ReCondition):
    def match(self, dataset: AbstractDataset):
        for task in dataset.tasks:
            if self.regex.search(task):
                return True
        return False


class RepositoryCondition(ReCondition):
    def match(self, dataset: AbstractDataset):
        return self.regex.search(dataset.repository.id) is not None

    def __repr__(self):
        return f"repository ~ {self.regex}"


class IDCondition(ReCondition):
    def match(self, dataset: AbstractDataset):
        return self.regex.search(dataset.id)

    def __repr__(self):
        return f"dataset ~ {self.regex}"


class TypeCondition(Condition):
    def __init__(self, typename: str):
        self.typename = typename

    def match(self, dataset: AbstractDataset):
        for ds in dataset.ancestors():
            if str(ds.__xpmtype__.identifier) == self.typename:
                return True

        return False



from pathlib import Path
from typing import Iterable, Iterator, Dict, Union
import importlib
import os
import hashlib
import logging
import inspect
import json
from abc import ABC, abstractmethod
from experimaestro import Config
import pkg_resources
from experimaestro.compat import cached_property
from experimaestro.mkdocs.metaloader import Module
from .utils import CachedFile, downloadURL
from .settings import UserSettings, Settings
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from datamaestro.definitions import AbstractDataset, DatasetWrapper


class Compression:
    @staticmethod
    def extension(definition):
        if not definition:
            return ""
        if definition == "gzip":
            return ".gz"

        raise Exception("Not handled compression definition: %s" % definition)


class Context:
    """
    Represents the application context
    """

    MAINDIR = Path(os.environ.get("DATAMAESTRO_DIR", "~/datamaestro")).expanduser()

    INSTANCE = None

    """Main settings"""

    def __init__(self, path: Path = None):
        assert not Context.INSTANCE

        Context.INSTANCE = self
        self._path = path or Context.MAINDIR
        self._dpath = Path(__file__).parents[1]
        self._repository = None
        # self.registry = Registry(self.datapath / "registry.yaml")
        self.keep_downloads = False
        self.traceback = False

        # Read global preferences
        self.settings = Settings.load(self._path / "settings.json")

        # Read user preferences
        path = Path("~").expanduser() / ".config" / "datamaestro" / "user.json"
        self.user_settings = UserSettings.load(path)

    @staticmethod
    def instance():
        if Context.INSTANCE is None:
            Context.INSTANCE = Context()
        return Context.INSTANCE

    @staticmethod
    def remote(host, pythonpath, datapath=None):
        """Create a remote context by connecting to a given host"""
        from experimaestro.rpyc import client

        client = client(hostname=host, pythonpath=pythonpath).__enter__()
        context = client.connection.modules.datamaestro.context.Context(datapath)
        return context

    @staticmethod
    def frompath(path: Path):
        context = Context.instance()

        class ContextManager:
            def __enter__(self):
                self.previous = Context.INSTANCE
                return context

            def __exit__(self, exc_type, exc_val, exc_tb):
                Context.INSTANCE = self.previous

        return ContextManager()

    @property
    def storepath(self):
        """Replaces the data path"""
        return self._path.joinpath("store")

    @property
    def datapath(self):
        return self._path.joinpath("data")

    @property
    def cachepath(self) -> Path:
        return self._path.joinpath("cache")

    @cached_property
    def repositorymap(self) -> Dict[str, "Repository"]:
        return {
            repository.basemodule(): repository
            for repository in self.repositories()
            if repository.basemodule() is not None
        }

    def repositories(self) -> Iterable["Repository"]:
        """Returns an iterator over repositories"""
        for entry_point in pkg_resources.iter_entry_points("datamaestro.repositories"):
            yield entry_point.load().instance()

    def repository(self, repositoryid):
        if repositoryid is None:
            return None

        entry_points = [
            x
            for x in pkg_resources.iter_entry_points(
                "datamaestro.repositories", repositoryid
            )
        ]
        if not entry_points:
            raise Exception("No datasets repository named %s", repositoryid)
        if len(entry_points) > 1:
            raise Exception(
                "Too many datasets repository named %s (%d)"
                % (repositoryid, len(entry_points))
            )
        return entry_points[0].load()(self)

    @property
    def running_test(self):
        return "PYTEST_CURRENT_TEST" in os.environ

    def datasets(self):
        """Returns an iterator over all files"""
        for repository in self.repositories():
            for dataset in repository:
                yield dataset

    def dataset(self, datasetid) -> "AbstractDataset":
        """Get a dataset by ID"""
        for repository in self.repositories():
            dataset = repository.search(datasetid)
            if dataset is not None:
                return dataset

        raise Exception("Dataset {} not found".format(datasetid))

    def downloadURL(self, url, size: int = None):
        """Downloads an URL

        Args:
            url (str): The URL to download
            size (str): The size if known (in bytes)
        """

        self.cachepath.mkdir(exist_ok=True)

        def getPaths(hasher):
            """Returns a cache file path"""
            path = self.cachepath.joinpath(hasher.hexdigest())
            urlpath = path.with_suffix(".url")
            dlpath = path.with_suffix(".dl")

            if urlpath.is_file():
                if urlpath.read_text() != url:
                    # TODO: do something better
                    raise Exception(
                        "Cached URL hash does not match. Clear cache to resolve"
                    )
            return urlpath, dlpath

        hasher = hashlib.sha256(json.dumps(url).encode("utf-8"))

        urlpath, dlpath = getPaths(hasher)
        urlpath.write_text(url)

        if dlpath.is_file():
            logging.debug("Using cached file %s for %s", dlpath, url)
        else:
            logging.info("Downloading %s", url)
            tmppath = dlpath.with_suffix(".tmp")

            downloadURL(url, tmppath, tmppath.is_file(), size=size)

            # Now, rename to original
            tmppath.rename(dlpath)

        return CachedFile(dlpath, keep=self.keep_downloads, others=[urlpath])

    def ask(self, question: str, options: Dict[str, str]):
        """Ask a question to the user"""
        print(question)  # noqa: T201
        answer = None
        while answer not in options:
            answer = input().strip().lower()
        return options[answer]


class ResolvablePath:
    """An object than can be resolved into a Path"""

    @staticmethod
    def resolve(context, path):
        if isinstance(path, ResolvablePath):
            return path(context)
        return Path(path)

    """Class that returns a path"""

    def __call__(self, context: Context) -> Path:
        raise NotImplementedError()


class DatafolderPath(ResolvablePath):
    def __init__(self, folderid, path):
        self.folderid = folderid
        self.path = path

    def __str__(self):
        return "datafolder-path({folderid}):{path}".format(**self.__dict__)

    def __call__(self, context: Context) -> Path:
        return Path(context.settings.datafolders[self.folderid]) / self.path


class Datasets(Iterable["AbstractDataset"]):
    """A set of datasets contained within a Python module"""

    def __init__(self, module: Module):
        """Initialize with a module"""
        self.module = module
        self._title = None
        self._description = None

    @property
    def id(self):
        return ".".join(self.module.__name__.split(".", 2)[2:])

    @property
    def title(self):
        self._getdoc()
        return self._title

    @property
    def description(self):
        self._getdoc()
        return self._description

    def _getdoc(self):
        if self._title is not None:
            return

        if not self.module.__doc__:
            self._title = ""
            self._description = ""
            return

        intitle = True
        title = []
        description = []
        for line in self.module.__doc__.split("\n"):
            if line.strip() == "" and intitle:
                intitle = False
            else:
                (title if intitle else description).append(line)

        self._title = " ".join(title)
        self._description = "\n".join(description)

    def __iter__(self) -> Iterable["AbstractDataset"]:
        from .definitions import DatasetWrapper
        from datamaestro.data import Base

        # Iterates over defined symbols
        for key, value in self.module.__dict__.items():
            # Ensures it is annotated
            if isinstance(value, DatasetWrapper):
                # Ensure it comes from the module
                if self.module.__name__ == value.t.__module__:
                    yield value
            elif (
                inspect.isclass(value)
                and issubclass(value, Base)
                and hasattr(value, "__dataset__")
            ):
                if self.module.__name__ == value.__module__:
                    yield value.__dataset__


class BaseRepository(ABC):
    """A repository groups a set of datasets and their corresponding specific
    handlers (downloading, filtering, etc.)"""

    def __init__(self, context: Context):
        self.context = context
        p = inspect.getabsfile(self.__class__)
        self.basedir = Path(p).parent

    @abstractmethod
    def __iter__(self) -> Iterator["AbstractDataset"]:
        ...

    def search(self, name: str):
        """Search for a dataset in the definitions"""
        for dataset in self:
            if name in dataset.aliases:
                return dataset

    @classmethod
    def instance(cls, context=None):
        try:
            return cls.__getattribute__(cls, "INSTANCE")
        except AttributeError:
            return cls(context if context else Context.instance())

    @classmethod
    def basemodule(cls):
        return cls.__module__

    @property
    def generatedpath(self):
        return self.basedir / "generated"

    @property
    def datapath(self):
        return self.context.datapath.joinpath(self.id)

    @property
    def extrapath(self):
        """Path to the directory containing extra configuration files"""
        return self.basedir / "data"


class Repository(BaseRepository):
    """(deprecated) Repository where datasets are located in __module__.config"""

    def __init__(self, context: Context):
        """Initialize a new repository

        :param context: The dataset main context
        :param basedir: The base directory of the repository
            (by default, the same as the repository class)
        """
        super().__init__(context)
        self.context = context
        self.configdir = self.basedir.joinpath("config")
        self.id = self.__class__.NAMESPACE
        self.name = self.id
        self.module = self.__class__.__module__
        self.__class__.INSTANCE = self

    @classmethod
    def version(cls):
        from pkg_resources import get_distribution, DistributionNotFound

        try:
            return get_distribution(cls.__module__).version
        except DistributionNotFound:
            return None

    def __repr__(self):
        return "Repository(%s)" % self.basedir

    def __hash__(self):
        return self.basedir.__hash__()

    def __eq__(self, other):
        assert isinstance(other, Repository)
        return self.basedir == other.basedir

    def datasets(self, candidate: str):
        """Returns the dataset candidates from a module"""
        try:
            module = importlib.import_module("%s.config.%s" % (self.module, candidate))
        except ModuleNotFoundError:
            return None
        return Datasets(module)

    def modules(self) -> Iterator["Module"]:
        """Iterates over all modules in this repository"""
        for _, fid, package in self._modules():
            try:
                module = importlib.import_module(package)
                yield Datasets(module)
            except Exception as e:
                import traceback

                traceback.print_exc()
                logging.error("Error while loading module %s: %s", package, e)

    def _modules(self):
        """Iterate over modules (without parsing them)"""
        for path in self.configdir.rglob("*.py"):
            try:
                relpath = path.relative_to(self.configdir)
                c = [p.name for p in relpath.parents][:-1][::-1]
                if path.name != "__init__.py":
                    c.append(path.stem)
                fid = ".".join(c)

                package = ".".join([self.module, "config", *c])

                yield self, fid, package
            except Exception as e:
                import traceback

                traceback.print_exc()
                logging.error("Error while reading definitions file %s: %s", path, e)

    def __iter__(self) -> Iterator["AbstractDataset"]:
        """Iterates over all datasets in this repository"""
        for datasets in self.modules():
            for dataset in datasets:
                yield dataset


def find_dataset(dataset_id: str):
    """Find a dataset given its id"""
    from .definitions import AbstractDataset

    return AbstractDataset.find(dataset_id)


def prepare_dataset(dataset_id: Union[str, "DatasetWrapper", Config]):
    """Find a dataset given its id and download the resources"""
    from .definitions import AbstractDataset, DatasetWrapper

    if isinstance(dataset_id, DatasetWrapper):
        ds = dataset_id
    elif isinstance(dataset_id, Config):
        ds = dataset_id.__datamaestro_dataset__
    else:
        ds = AbstractDataset.find(dataset_id)

    return ds.prepare(download=True)


def get_dataset(dataset_id: str):
    """Find a dataset given its id"""
    from .definitions import AbstractDataset

    ds = AbstractDataset.find(dataset_id)
    return ds.prepare(download=False)



import logging
from pathlib import Path
import os

from datamaestro import AbstractDataset
from datamaestro.download import Download


class List(Download):
    """Download multiple files or directories given by a list"""

    def __init__(self, dataset: AbstractDataset, definition: object):
        super().__init__(dataset, definition)
        self.list = self.definition

    def download(self, destination):
        logging.info("Downloading %d items", len(self.list))
        for key, value in self.list.items():
            if not key.startswith("__"):
                handler = Download.find(self.dataset, value)
                destpath = handler.path(destination, key)
                handler.download(destpath)

    def files(self, destpath):
        """Set the list of files"""
        r = {}
        for key, value in self.list.items():
            if not key.startswith("__"):
                handler = Download.find(self.dataset, value)
                r[key] = handler.files(destpath)
        return r


class Datasets(Download):
    """Use links to dataset files"""

    def __init__(self, dataset: AbstractDataset, definition: object):
        super().__init__(dataset, definition)
        self.list = self.definition

    def download(self, destination):
        destination.mkdir(exist_ok=True, parents=True)

        for key, value in self.list.items():
            if not key.startswith("__"):
                files = value.prepare().files

                if isinstance(files, Path):
                    if not files.is_dir():
                        raise AssertionError(
                            "Dataset path is not a directory: %s", files
                        )
                    path = destination / key
                    if not path.exists():
                        if path.is_symlink():
                            logging.warning("Path %s is symlink", path)
                        else:
                            os.symlink(files, path)
                elif len(files) > 1:
                    raise NotImplementedError()

    def files(self, destpath):
        """Set the list of files"""
        return destpath



from pathlib import Path


class Transform:
    @staticmethod
    def createFromPath(path: Path):
        if path.suffix == ".gz":
            from .compress import Gunzip

            return Gunzip()
        return Identity()

    def path(self, path):
        return path

    def __call__(self, input):
        raise NotImplementedError(
            "__call__ should be implemented in subclass %s" % type(self)
        )


class Identity(Transform):
    def __call__(self, fileobj):
        return fileobj


class TransformList(Transform):
    def __init__(self, *list):
        self.list = list

    def append(self, item):
        self.list.append(item)

    def path(self, path):
        for item in self.list:
            path = item.path(path)
        return path

    def __call__(self, fileobj):
        for item in self.list:
            fileobj = item(fileobj)
        return fileobj



from abc import ABC, abstractmethod
from struct import Struct
from typing import TYPE_CHECKING
from . import File, Base

if TYPE_CHECKING:
    import numpy as np


class Tensor(Base, ABC):
    @abstractmethod
    def data(self) -> "np.ndarray":
        """Returns the tensor in numpy format"""
        pass


class IDX(Tensor, File):
    """IDX File format

    The IDX file format is a simple format for vectors and multidimensional
    matrices of various numerical types.

    The basic format is:

    magic number size in dimension 0 size in dimension 1 size in dimension 2
    ..... size in dimension N data

    The magic number is an integer (MSB first). The first 2 bytes are always 0.

    The third byte codes the type of the data: 0x08: unsigned byte 0x09: signed
    byte 0x0B: short (2 bytes) 0x0C: int (4 bytes) 0x0D: float (4 bytes) 0x0E:
    double (8 bytes)

    The 4-th byte codes the number of dimensions of the vector/matrix: 1 for
    vectors, 2 for matrices....

    The sizes in each dimension are 4-byte integers (MSB first, high endian,
    like in most non-Intel processors).

    The data is stored like in a C array, i.e. the index in the last dimension
    changes the fastest.
    """

    MAGIC_NUMBER = Struct(">HBB")
    DIM = Struct(">I")

    def data(self) -> "np.ndarray":
        """Returns the tensor"""
        import numpy as np

        with self.open("rb") as fp:
            zero, magic, size = IDX.MAGIC_NUMBER.unpack_from(
                fp.read(IDX.MAGIC_NUMBER.size)
            )
            if zero != 0:
                raise IOError("File format not IDX (the two first bytes are not zero)")

            if magic == 0x08:
                dtype = ">B"
            elif magic == 0x09:
                dtype = ">b"
            else:
                raise IOError("Magic number {} not recognized in IDX file", magic)

            shape = [IDX.DIM.unpack_from(fp.read(IDX.DIM.size))[0] for i in range(size)]

            size = np.prod(shape)
            # Could use np.fromfile... if it were not broken
            # see https://github.com/numpy/numpy/issues/7989
            data = np.frombuffer(fp.read(), dtype=dtype, count=size)
            data = data.reshape(shape, order="C")
        return data
